{$A+,B-,E-,F-,G-,I-,N-,O-,V-,X- }

{$IFDEF DEBUG }
  {$R+,S+ }
{$ELSE }
  {$D-,L-,R-,S- }
{$ENDIF }

Unit EmulM2;

Interface
uses VdtDisp;


const ScreenLong = 40;
      ScreenHigh = 25;  { ligne 0 speciale }

      LongPCode = 10;

      CoulPC : array[0..7] of byte = (0,2,4,6,1,3,5,7);

      { EMOP = EMul OPration }

      EMOP_NIL            = 0;     { rien de particulier }
      EMOP_CHGSTATUSFONC  = 1;     { nouveau mode de fonctionnement }
      EMOP_ACKRESET       = 2;     { Ack Reset et equivalent }

      EMOP_ROLLON         = 32;    { passage en mode rouleau }
      EMOP_ROLLOFF        = 33;    { passage en mode page }
      EMOP_RESET          = 34;    { Reset minitel }
      EMOP_CONNEXION      = 35;    { Connexion }
      EMOP_DECONNEXION    = 36;    { Deconnexion }
      EMOP_CURON          = 37;    { Curseur allum }
      EMOP_CUROFF         = 38;    { Curseur eteind }
      EMOP_RET            = 39;    { retournement minitel }
      EMOP_RETINV         = 40;    { retournement inverse }
      EMOP_ENQROM         = 41;    { Id ROM }
      EMOP_PROCON         = 42;    { activation correction d'erreur }
      EMOP_PROCOFF        = 43;    { desactivation ... }
      EMOP_POSCUR         = 44;    { demande position curseur }
      EMOP_IDEN1          = 45;    { Entete bloc 1 }
      EMOP_IDEN2          = 46;    { Entete bloc 2 }
      EMOP_ENQRAM1        = 47;    { contenu ram 1 }
      EMOP_ENQRAM2        = 48;    { contenu ram 2 }
      EMOP_TELEINFO       = 49;    { passage en standard teleinfo }
      EMOP_MIXTE1         = 50;    { passage mode mixte, jeu americain }
      EMOP_MIXTE2         = 51;    { passage mode videotex }
      EMOP_REPSTATUSFONC  = 52;    { reponse a la demande de status }
      EMOP_OPPO           = 53;    { retournement pour opposabilit }
      EMOP_OPPORE         = 54;    { passage mode oppos  esclave }
      EMOP_MINUSCULE      = 55;
      EMOP_MAJUSCULE      = 56;
      EMOP_MSKON          = 57;    { Masque plein ecran ON }
      EMOP_MSKOFF         = 58;    { Masque plein ecran OFF }


type EmulCfg = record
         EmulTrans     : boolean;  { Emulation des transparences }
         VisuelConvert : boolean;  { Conversion visuelle pour pc }
         RTIC          : boolean;  { M1B de type RTIC ?          }
         InsLowOnly    : boolean;  { en insertion, le bas bouge seulement (M2) }
         C15Clr        : boolean;  { si le code 15 annule le soul latent (M2) }
       end;

     EmulRdy = Record
       CurOFFON     : Boolean;  { TRUE=>on sait si CurOn ou OFF }
      End;

     EmulSitu = Record
       X,Y        : byte;     { coord du curseur (0..39,0..24) }
       G0         : Boolean;  { FALSE=> jeu G'0 actif !!! }
       G1         : Boolean;  { FALSE=> jeu G'1 actif !!! }
       Att        : word;     { attribut courant du curseur }
       Latent     : word;     { attribut latent du curseur (dlimiteur) }
       CurPos     : Boolean;  { TRUE=>on connait la pos du cur }
      End;

     EmulStat = record
         LastCar    : char;     { dernier caractre reu }
         LastNotDRCS: Boolean;  { Exception a la propagation DRCS ??? }
         MdIns      : boolean;  { mode insertion en criture }
         MdPage     : boolean;  { mode page ( sinon rouleau ) }
         CurON      : boolean;  { etat du curseur allum ? }
         nG         : Byte;     { njeu DRCS Tlchargeable (0 ou 1) }
         MskEcran   : Boolean;  { TRUE=>Masque plein ecran }
       end;

       (*
     EmulSave = record
         Use        : boolean;
         X,Y        : byte;
         Att        : word;
         Latent     : word;
         CurPos     : Boolean;
       end;
     *)

     EmulAiguillage = (NUL,REP,SEP,SS2,SS2_M,SS3,US,ESC,FILT,
                       PRO1,PRO2,PRO3,ISO_6429,ISO_2022,TRANS_MODE,
                       JEU_DRCS,TELE_DRCS);

type PRO_Rec = record
         Par1, Par2, Par3 : word;               { parametre aiguillage }
         Case Integer Of
          0 : ( PCode : array[0..LongPCode-1] of byte ); { parametre videotex }
          1 : ( Par4, Par5, Par6, Par7 : Word );
       end;

     EmulRec = record
         Aig   : EmulAiguillage; { aiguillage }
         Cfg   : EmulCfg;
         Stat  : EmulStat;
         Situ  : EmulSitu;
         Saved : Boolean;
         Save  : EmulSitu;
         Rdy   : EmulRdy;
         Ram   : PVDispRam;
         DRCS  : PVDispRamDRCS;
         Pro   : PRO_Rec;
       end;

     PEmulRec = ^EmulRec;


procedure InitEmul( var Em : EmulRec; Ram : PVDispRam; DRCS : PVDispRamDRCS );
procedure SetDefaultEmul( var Cfg : EmulCfg );
Function  Emul( var Em : EmulRec; Car : char ) : Byte;


Type EmulScrollProc = Record
       Case Integer Of
        0 : ( Proc : Procedure ( Ram : PVDispRam; y : Byte; Down : Boolean ) );
        1 : ( Ptr : Pointer );
      End;

Const ScrollProc : EmulScrollProc = ( Ptr : NIL );


implementation


Function  RamScroll( Ram : PVDispRam; y : Byte; Down : Boolean ) : Boolean;
Begin
 If ScrollProc.Ptr<>NIL Then
  Begin
   ScrollProc.Proc(Ram,y,Down);
   RamScroll := TRUE;
  End
 Else RamScroll := FALSE;
End;




const DefaultCfg : EmulCfg = (
          EmulTrans     : TRUE;    { Emulation des transparences }
          VisuelConvert : TRUE;    { Conversion visuelle pour pc }
          RTIC          : TRUE;    { M1B de type RTIC ?          }
          InsLowOnly    : FALSE;   { car M1B                     }
          C15Clr        : FALSE    { car emulateur M1B ...       }
        );

const Trait      = 7;
      Fond       = 56;
      Taille     = 192;
      Inverse    = 256;
      Stable     = 512;
      Disjoint   = 1024;
      Souligne   = 1024;
      Mask       = 2048;

      Graphik    = 8192;

      Delim      = 32768;

      Complement = 176;
      Haut       = 64;
      Large      = 128;

      AttNothing = Trait+Stable+Graphik;
      AttInit    = Trait+Stable;


      CarError     = 168;
      CarErrorM1B  = 127;

      RIEN        = 250;
      CarGraphik  = 240;



const { #25#x -> IBM o x<>65,66,67,72,75 }
      TabExt : array[32..127] of char =
        ('_','_','_','','$','_','#',#21,'_','_','_','_',#27,#24,#26,#25,
         '','','_','_','_','_','_','_','','_','_','_','','','','_',
         '_','_','_','_','_','_','_','_','_','_','_','_','_','_','_','_',
         '_','_','_','_','_','_','_','_','_','_','_','_','_','_','_','_',
         '_','_','_','_','_','_','_','_','_','_','','_','_','_','_','_',
         '_','_','_','_','_','_','_','_','_','_','','','_','_','_','_');

      TabAccent_A : array[65..72] of char = ('','a','','a','a','a','a','');
      TabAccent_E : array[65..72] of char = ('','','','e','e','e','e','');
      TabAccent_I : array[65..72] of char = ('i','i','','i','i','i','i','');
      TabAccent_O : array[65..72] of char = ('o','o','','o','o','o','o','');
      TabAccent_U : array[65..72] of char = ('','u','','u','u','u','u','');

function TraducteurTwo( CarOne, Car : char ) : char;
{ attention: dataOne=[65..72] ou 75 sinon ca plante! }
begin
  if CarOne=#75 then
    begin
      if Car='c' then Car := ''
      else if Car='C' then Car := '';
    end
  else
    case Car of
      'a' : Car := TabAccent_A[byte(CarOne)];
      'e' : Car := TabAccent_E[byte(CarOne)];
      'i' : Car := TabAccent_I[byte(CarOne)];
      'o' : Car := TabAccent_O[byte(CarOne)];
      'u' : Car := TabAccent_U[byte(CarOne)];
    end;
  TraducteurTwo := Car;
end;

procedure CurChange( var Em : EmulRec );
begin
end;


procedure ScrUp( var Em : EmulRec; Y : byte ); { monte l'ecran }
{ Y de 1..24 ( pas 0 ! ) }
var Ram : PVDispRam;
    i   : word;
begin
  if (Y>0) and (Y<ScreenHigh-1) then
    begin
      Ram := Em.Ram;
      move(Ram^[Y+1],Ram^[Y],sizeof(Ram^[0])*(ScreenHigh-(Y+1))); { +1 ... }
      for i := 0 to ScreenLong-1 do
        with Ram^[ScreenHigh-1,i] do
          begin
            Car := #0;
            Att := AttNothing;
          end;
      If Not(RamScroll(Ram,y,FALSE)) Then VDRamChange(Ram,0,Y,ScreenLong,ScreenHigh-Y);
    end;
end;

procedure ScrDown( var Em : EmulRec; Y : byte ); { descend l'ecran }
{ Y de 1..24 ( pas 0 ! ) }
var Ram : PVDispRam;
    i   : word;
begin
  if (Y>0) and (Y<ScreenHigh-1) then
    begin
      Ram := Em.Ram;
      move(Ram^[Y],Ram^[Y+1],sizeof(Ram^[0])*(ScreenHigh-(Y+1))); { +1 ... }
      for i := 0 to ScreenLong-1 do
        with Ram^[Y,i] do
          begin
            Car := #0;
            Att := AttNothing;
          end;
      If Not(RamScroll(Ram,y,TRUE)) Then VDRamChange(Ram,0,Y,ScreenLong,ScreenHigh-Y);
    end;
end;

procedure ClrLigneMntl( var Em : EmulRec; Y, Xstart, Lim : byte);
var i   : word;
    Ram : PVDispRam;
begin
  Ram := Em.Ram;
  for i := XStart+1 to Lim do
    with Ram^[Y,i] do
      begin
        Car := #0;
        Att := AttNothing;
      end;
  InRam(Ram,XStart,Y,#0,AttNothing);

  VDRamChange(Ram,XStart+1,Y,Lim-XStart,1);
end;

procedure ClrScrMntl( var Em : EmulRec; Y, Lim : byte );
var i : word;
begin
  for i := Y to Lim do ClrLigneMntl(Em,i,0,ScreenLong-1);
end;

procedure SupMntl( var Em : EmulRec; X, Y, Nbre : byte);
var Ram : PVDispRam;
    i   : word;
begin
  if X+Nbre<ScreenLong then
    begin
      Ram := Em.Ram;
      for i := X+Nbre to ScreenLong-1 do
        InRam(Ram,i-Nbre,Y,Ram^[Y,i].Car,Ram^[Y,i].Att);
      X := ScreenLong-Nbre;
    end;
  ClrLigneMntl(Em,Y,X,ScreenLong-1) { toutes la ligne a partir de X y passe ! }
end;

procedure InsMntl( var Em : EmulRec; X, Y, Nbre : byte );
var Ram : PVDispRam;
    i   : word;
begin
  if (X+Nbre)<ScreenLong then
    begin
      Ram := Em.Ram;
      move(Ram^[Y,X],Ram^[Y,X+Nbre],(ScreenLong-(X+Nbre))*sizeof(Ram^[0,0]));
      VDRamChange(Ram,X+Nbre,Y,ScreenLong-(X+Nbre),1);
    end
  else Nbre := ScreenLong-x;
  ClrLigneMntl(Em,Y,X,(X+Nbre)-1);
end;

{ ---------------------------------------------------------------- }

procedure SetDefaultEmul( var Cfg : EmulCfg );
begin
  Cfg := DefaultCfg;
end;

const DefaultStat : EmulStat = (
          LastCar : #0;
          LastNotDRCS : FALSE;
          MdIns   : FALSE;
          MdPage  : TRUE;
          CurON   : FALSE;
          nG      : 0;
          MskEcran: FALSE
        );

      DefaultSitu : EmulSitu = (
          X       : 0;
          Y       : 1;
          G0      : TRUE;
          G1      : TRUE;
          Att     : AttInit;
          Latent  : 0;
          CurPos  : FALSE
        );

procedure InitEmul( var Em : EmulRec; Ram : PVDispRam; DRCS : PVDispRamDRCS );
begin
  FillChar(Em,sizeof(Em),0);
  Em.Cfg  := DefaultCfg;
  Em.Ram  := Ram;
  Em.DRCS := DRCS;
  EM.Stat := DefaultStat;
  EM.Situ := DefaultSitu;
  EM.Aig  := NUL;
end;

procedure CurHome( var Em : EmulRec ); { code 12 ou 30 }
begin
  Em.Situ.X      := 0;
  Em.Situ.Y      := 1;
  Em.Situ.Att    := AttInit;
  Em.Situ.Latent := 0;
  Em.Situ.CurPos := TRUE;

  Em.Saved       := FALSE;

end;



function CorrigeCar( var Em : EmulRec; Car : char; Att : word ) : char;
begin
  { corrige Car }
  If Att And DRCS=0 Then { corrige si pas DRCS ! }
   if Att and Graphik=0 then
     begin { texte }
       if Em.Cfg.VisuelConvert then
         case Car of
            #96 : Car := #196;
           #124 : Car := #179;
         end;
     end
   else
    begin { graphik }
     if (Car>=#64) and (Car<=#94) then inc(Car,32)
     else if Car=#127 then Car := #95;
    end;
  CorrigeCar := Car;
end;

procedure PropAtt( Y : byte;
                   Car : char; var Att, Latent : word;
                   PrecAtt : word; CanBeDelim : boolean );
begin
  if Att and Graphik=0 then
    begin { mode texte }
      if CanBeDelim and (Car=#32) and (Latent<>0) then
        begin { en texte, qu'un seul est latent, et tous le deviennent }
          Att := Att or Delim;
          Latent := 0;
        end
      else
       Begin
        { le soulignement n'est pas propag si le prcdent est graphik... }

        If (PrecAtt And Graphik)<>0 Then PrecAtt := PrecAtt And Not(Souligne);

        Att := (Att and not(Fond+Souligne+Mask+Delim)) or
               (PrecAtt and (Fond+Souligne+Mask));
       End;
    end
  else
    begin { mode graphik }
      if CanBeDelim and (Car=#32) and (Latent and Mask<>0) then
        begin
          Att := Att or Delim;
          Latent := 0;
        end
      else
        begin
          Att := (Att and not(Mask+Delim)) or (PrecAtt and Mask);
          Latent := Latent and not(Fond+Souligne);
        end;
    end;
  if (Y<=1) and
     (Att and Taille=Taille) then Att := Att and not(Taille);
end;


procedure CurMove( var Em : EmulRec; var X, Y : byte; Att : word );
begin
  { deplace le curseur ... }
  if Att and Large<>0 then inc(X,2)
  else inc(X);
  if X>ScreenLong-1 then
    if Y>0 then
      begin
         X := 0;
         if Att and Haut<>0 then
           begin
             inc(Y,2);
             if Y>ScreenHigh-1 then
               if Em.Stat.MdPage then Y := 2
               else
                 begin
                   ScrUp(Em,1);
                   if Y>ScreenHigh then ScrUp(Em,1);
                   Y := ScreenHigh-1;
                 end;
           end
         else
           begin
             inc(Y);
             if Y>ScreenHigh-1 then
               if Em.Stat.MdPage then Y := 1
               else
                 begin
                   ScrUp(Em,1);
                   Y := ScreenHigh-1;
                 end;
           end;
      end
    else X := ScreenLong-1; { dbordement impossible en ligne 0 }
end;

procedure InsertAtt( var Em : EmulRec; X, Y : byte; Att : word );
begin
  if (Att and Haut<>0) and (Y>1) and (Em.Cfg.InsLowOnly=FALSE) then
    begin
      if (Att and Large<>0) and (X<ScreenLong-1) then InsMntl(Em,X,Y-1,2)
      else InsMntl(Em,X,Y-1,1);
    end;
  if (Att and Large<>0) and (X<ScreenLong-1) then InsMntl(Em,X,Y,2)
  else InsMntl(Em,X,Y,1);
end;


Function AddDRCS( Var Em : EmulRec; Att : Word ) : Word;
Begin
 If Att And Graphik=0 Then
  Begin
   If Not(Em.Situ.G0) Then Att := Att Or DRCS;
  End
 Else
  Begin
   If Not(Em.Situ.G1) Then Att := Att Or DRCS;
  End;

 AddDRCS := Att;
End;

procedure NewCar( var Em : EmulRec; Car : char; NotDRCS : Boolean );
var Ram     : PVDispRam;
    PrecAtt : word;
    Att     : word;
    X, Y    : byte;
    Latent  : word;
begin
  Ram := Em.Ram;
  if Em.Situ.X>0 then PrecAtt := Ram^[Em.Situ.Y,Em.Situ.X-1].Att
  else PrecAtt := AttNothing;
  Att := Em.Situ.Att;
  If Not(NotDRCS) Then Att := AddDRCS(Em,Em.Situ.Att);
  X   := Em.Situ.X;
  Y   := Em.Situ.Y;
  Latent := Em.Situ.Latent;

  Em.Stat.LastCar     := Car; { toujours avant la correction ! }
  Em.Stat.LastNotDRCS := NotDRCS;

  Car := CorrigeCar(Em,Car,Att);

  PropAtt(Y,Car,Att,Latent,PrecAtt,TRUE);

  if (Em.Stat.MdIns) and (Y>0) then InsertAtt(Em,X,Y,Att);

  { on doit tenir compte des insertions en criture ! }

  WriteVDRam(Em.Ram,X,Y,Car,Att); { place en Ram }
  CurMove(Em,X,Y,Att);

  Em.Situ.X := X;
  Em.Situ.Y := Y;
  Em.Situ.Att := (Em.Situ.Att and not(Taille)) or (Att and Taille);
  Em.Situ.Latent := Latent;
  CurChange(Em);
end;

procedure CAN( var Em : EmulRec );
var Att     : word;
    PrecAtt : word;
    i       : word;
    Ram     : PVDispRam;
    Latent  : word;
begin
  Ram := Em.Ram;
  Att := AddDRCS(Em,Em.Situ.Att);
  Latent := Em.Situ.Latent;
  if Em.Situ.X>0 then PrecAtt := Ram^[Em.Situ.Y,Em.Situ.X-1].Att else PrecAtt := AttNothing;

  PropAtt(Em.Situ.Y,#32,Att,Latent,PrecAtt,FALSE); { ne peut etre delimiteur }

  for i := Em.Situ.X to ScreenLong-1 do
    begin
      DirectVDRam(Ram,i,Em.Situ.Y,#32,Att);
      if (Att and Large<>0) and (i<ScreenLong-1) then inc(i);
    end;
  Em.Situ.Att := (Em.Situ.Att and not(Taille)) or (Att and Taille);
  Em.Situ.Latent := Latent;
end;

procedure QuitLine0( var Em : EmulRec );
begin
  if Em.Saved then
    begin
      Em.Situ  := Em.Save;
      Em.Saved := FALSE;
    end
  else
    begin { y a une erreur quelquepart ... }
      Em.Situ := DefaultSitu;
    end;
end;

procedure Repetition( var Em : EmulRec; How : byte );
var Ram     : PVDispRam;
    PrecAtt : word;
    Att     : word;
    X, Y    : byte;
    Latent  : word;
    Car     : char;
begin
  if Em.Stat.LastCar<>#0 then
    begin
      Ram := Em.Ram;
      X   := Em.Situ.X;
      Y   := Em.Situ.Y;
      Latent := Em.Situ.Latent;

      Att := Em.Situ.Att;
      If Not(Em.Stat.LastNotDRCS) Then Att := AddDRCS(Em,Em.Situ.Att);

      Car := CorrigeCar(Em,Em.Stat.LastCar,Att);

      while How>0 do
        begin
          if X>0 then PrecAtt := Ram^[Y,X-1].Att
          else PrecAtt := AttNothing;
          PropAtt(Y,Car,Att,Latent,PrecAtt,FALSE);

          if (Em.Stat.MdIns) and (Y>0) then InsertAtt(Em,X,Y,Att);
          { on doit tenir compte des insertions en criture ! }

          if How>1 then DirectVDRam(Ram,X,Y,Car,Att) { <- on va + vite }
          else WriteVDRam(Ram,X,Y,Car,Att); { dernire ecriture }

          CurMove(Em,X,Y,Att);

          dec(How);
        end;

      Em.Situ.X := X;
      Em.Situ.Y := Y;

      Em.Situ.Att := (Em.Situ.Att and not(Taille)) or (Att and Taille);
      Em.Situ.Latent := Latent;
      CurChange(Em);
    end;
end;


Procedure ResetMinitel( Var Em : EmulRec );
Begin
 { eteindre curseur }

 Em.Stat.CurOn := FALSE;
 Em.Rdy.CurOFFON := TRUE;

 { mode page }

 Em.Stat.MdPage := TRUE;

 { changement de l'att curseur ! }

 Em.Situ.Att    := AttInit;
 Em.Situ.Latent := 0;

 { plus d'insertion }

 Em.Stat.MdIns  := FALSE;

 { Jeu 0 }
 Em.Stat.nG     := 0;

 { G standard }

 Em.Situ.G0     := TRUE;
 Em.Situ.G1     := TRUE;

 { Plus de masquage plein ecran }

 Em.Stat.MskEcran := FALSE;
End;


Function  EmulCtrl( Var Em : EmulRec; Car : Char ) : Byte;
Begin
 EmulCtrl := EMOP_NIL;

 case Car of
   #05 : EmulCtrl := EMOP_ENQRAM1;    { Contenu RAM 1 }
   #08 : { curseur vers la gauche }
      begin
        if Em.Situ.X=0 then
          begin
            if Em.Situ.Y>0 then
              begin
                Em.Situ.X := ScreenLong-1;
                dec(Em.Situ.Y);
                if Em.Situ.Y=0 then
                  if Em.Stat.MdPage then Em.Situ.Y := ScreenHigh-1
                  else
                    begin
                      Em.Situ.Y := 1;
                      ScrDown(Em,1);
                    end;
              end;
              { sinon, en ligne 0, on ne bouge pas ! }
          end
        else dec(Em.Situ.X);
        CurChange(Em);
      end;
   #09 : { curseur vers la droite }
      begin
        inc(Em.Situ.X);
        if Em.Situ.X>=ScreenLong then
          begin
            if Em.Situ.Y>0 then
              begin
                Em.Situ.X := 0;
                inc(Em.Situ.Y);
                if Em.Situ.Y>=ScreenHigh then
                  if Em.Stat.MdPage then Em.Situ.Y := 1
                  else
                    begin
                      Em.Situ.Y := ScreenHigh-1;
                      ScrUp(Em,1);
                    end;
              end
            else Em.Situ.X := ScreenLong-1;
              { pas de dbordement en ligne 0 }
          end;
        CurChange(Em);
      end;
   #10 : { curseur vers le bas }
      begin
        if Em.Situ.Y>0 then
          begin
            inc(Em.Situ.Y);
            if Em.Situ.Y>=ScreenHigh then
              if Em.Stat.MdPage then Em.Situ.Y := 1
              else
                begin
                  Em.Situ.Y := ScreenHigh-1;
                  ScrUp(Em,1);
                end;
          end
        else { on quitte la ligne 0 }
          QuitLine0(Em);
        CurChange(Em);
      end;
   #11 : { curseur vers le haut }
      if Em.Situ.Y>0 then
        begin
          dec(Em.Situ.Y);
          if Em.Situ.Y=0 then
            if Em.Stat.MdPage then Em.Situ.Y := ScreenHigh-1
            else
              begin
                Em.Situ.Y := 1;
                ScrDown(Em,1);
              end;
          CurChange(Em);
        end;
      { ignore en ligne 0 }
   #12 : { efface ecran }
        begin
          ClrScrMntl(Em,1,ScreenHigh-1);
          CurHome(Em);
        end;
   #13 : { retour chariot }
        begin
          Em.Situ.X := 0;
          CurChange(Em);
        end;
   #14 : { mode graphik }
        begin
          Em.Situ.Att    := (Em.Situ.Att and not(Disjoint+Inverse+Taille))
                            or Graphik;
          Em.Situ.Latent := Em.Situ.Latent and not(Disjoint);
        end;
   #15 : { mode texte }
        begin
          if Em.Cfg.C15Clr then
            begin
              Em.Situ.Latent := Em.Situ.Latent and not(Souligne);
            end
          else
            begin
              if Em.Situ.Att and Graphik<>0 then
                Em.Situ.Latent := Em.Situ.Latent and not(Souligne);
                { passage graphik->texte, le soulignement latent }
                { est supprim }
            end;
          Em.Situ.Att := Em.Situ.Att and
                         not(Souligne+Inverse+Taille+Graphik);
        end;
   #17 : Begin
          Em.Stat.CurOn := TRUE;
          Em.Rdy.CurOFFON := TRUE;
          EmulCtrl := EMOP_CURON;
         End;
   #18 : Em.Aig := REP; { repetition }
   #19 : Em.Aig := SEP; { Sep (fonction) }
   #20 : Begin
          Em.Stat.CurOn := FALSE;
          Em.Rdy.CurOFFON := TRUE;
          EmulCtrl := EMOP_CUROFF;
         End;
   #24 : CAN(Em); { propage des espaces jusqu'a la fin de ligne, sans }
                  { deplacement curseur. Ce n'est pas un dlimiteur!  }
   #22,#25 : { extension alpha }
        { ce code est ignore en mode graphik !! }
        if Em.Situ.Att and Graphik=0 then Em.Aig := SS2;
   #26 : { symbole d'erreur de reception }
         NewCar(Em,char(CarErrorM1B),TRUE);
   #27 : Em.Aig := ESC; { ohlala! }
   #29 : Em.Aig := SS3; { SS3,X }
   #30 : CurHome(Em); { curseur de retour en position initiale }
   #31 : begin { positionnement du curseur }
           Em.Aig      := US;
           Em.Pro.Par3 := 0;
         end;
 end; { case of }
End;


Procedure AccuTele( Var Em : EmulRec; Car : Char );
Var Mat : PMinMatrice;
    o   : Byte;
Begin
 If Em.Pro.Par3=0 Then
  Begin
   Em.Pro.Par5 := 0;  { n bit }
   Em.Pro.Par6 := 0;  { noctet matrice }
  End;

 Mat := @Em.DRCS^[Em.Stat.nG,Em.Pro.Par2-Byte('!')];

 Case Em.Pro.Par3 Of
  0, 4, 8, 12 : ;
  1, 5, 9, 13 : Begin
                 o := ((Em.Pro.Par4 And 63) Shl 2) Or (Byte(Car) Shr 4) And 3;
                 Mat^[Em.Pro.Par6] := o;
                 Inc(Em.Pro.Par6);
                End;
  2, 6, 10 : Begin
              o := ((Em.Pro.Par4 And 15) Shl 4) Or ((Byte(Car) Shr 2) And 15);
              Mat^[Em.Pro.Par6] := o;
              Inc(Em.Pro.Par6);
             End;
  3, 7, 11 : Begin
              o := ((Em.Pro.Par4 And 3) Shl 6) Or (Byte(Car) And 63);
              Mat^[Em.Pro.Par6] := o;
              Inc(Em.Pro.Par6);
             End;
 End;

 Em.Pro.Par4 := Byte(Car);

 Inc(Em.Pro.Par3);
 If Em.Pro.Par3>=14 Then
  Begin
   Inc(Em.Pro.Par2);
   Em.Pro.Par1 := 0;
   Em.Pro.Par3 := 0;
  End;
End;


Procedure CompleteTele( Var Em : EmulRec );
Begin
 While Em.Pro.Par1<>0 Do AccuTele(Em,#$40);
End;


Function  Emul( var Em : EmulRec; Car : char ) : Byte;
begin
 Emul := EMOP_NIL;

 Car := char(byte(Car) and 127);
 if (Car=#0) then exit;

 If (Car<#32) And
    Not(Em.Aig In [TRANS_MODE,TELE_DRCS]) Then
  Begin
   Em.Aig := NUL;
   Emul := EmulCtrl(Em,Car);
  End
 Else
  Case Em.Aig Of
   NUL : NewCar(Em,Car,FALSE);

   REP :
       begin { repetion }
         if Car>=#64 then Repetition(Em,byte(Car)-64);
         Em.Aig := NUL;
       end;
   SEP :
       begin { sep }
         if Car In [#$53,   { deco/connexion }
                    #$59,   { phase deco/connexion }
                    #$5E,   { Reset }
                    #$70,   { Passage au mode Mixte }
                    #$71]   { Passage au mode Videotex }
                  Then
           begin { reset }
            ResetMinitel(Em);
            Emul := EMOP_ACKRESET;
           end
         Else If Car=#$56 Then Emul := EMOP_CHGSTATUSFONC;

         Em.Aig := NUL;
       end;
   SS2 : { code alpha etendu }
       if Car in [#65,#66,#67,#72,#75] then
         begin
           Em.Pro.Par1 := byte(Car); { sauvegarde }
           Em.Aig      := SS2_M;
         end
       else
         begin
           if not(Car in [#68..#71,#74,#77..#79]) then
              NewCar(Em,TabExt[byte(Car)],TRUE);
           { en effet, certain code filtre au lieu d'emettre un 'Inconnu' }
           Em.Aig := NUL;
         end;
   SS2_M : { code avec accent ! }
           { "Car" contient le caractere et Par1 le type d'accent }
       begin
        If Em.Situ.G0 Then NewCar(Em,TraducteurTwo(char(Em.Pro.Par1),Car),FALSE)
        Else NewCar(Em,Car,FALSE);  { ne traduit pas... }
        Em.Aig := NUL;
       end;
   SS3 : Em.Aig := NUL; { SS3, X  ignore ... }
   US : begin { positionnement curseur }
         if Em.Pro.Par3=0 then
           begin
            Em.Pro.Par1 := byte(Car); { garde }
            Em.Pro.Par3 := 1;         { pour la deuxieme coord }
           end
         else
           begin { on a les coords }
             Em.Pro.Par2 := byte(Car);
             If (Em.Pro.Par1=$23) And (Em.DRCS<>NIL) Then
              Begin { Cas spcial !! }
               If Car=#$20 Then Em.Aig := JEU_DRCS { change jeu telechargeable }
               Else Em.Aig := TELE_DRCS;   { entte telechargement ... }

               Em.Pro.Par1 := 0;
               Em.Pro.Par3 := 0;
              End
             Else
              Begin
                   { Par1:ligne  Par2:colonne }
               if (Em.Pro.Par1>=48) and (Em.Pro.Par1<=50) and
                  (Em.Pro.Par2>=48) and (Em.Pro.Par2<=57) then
                 begin { traduction car fonction CEPT }
                   Em.Pro.Par1 := 64+(Em.Pro.Par1-48)*10+(Em.Pro.Par2-48);
                   Em.Pro.Par2 := 65;
                 end;
               if (Em.Pro.Par1>=64) and (Em.Pro.Par1<=88) and
                  (Em.Pro.Par2>=65) and (Em.Pro.Par2<=104) then
                 begin { coord bonne }
                   if Em.Pro.Par1=64 then
                     begin { ligne Zero!! }
                       if Em.Situ.Y>0 then
                         begin { on va en ligne 0 ... }
                               { on sauve l'etat courant }
                           Em.Saved := TRUE;
                           Em.Save  := Em.Situ;

                           Em.Situ.G0 := TRUE;  { Normal ... }
                           Em.Situ.G1 := TRUE;
                         end;
                     end
                   else
                     begin { on va pas en ligne 0 }
                       if Em.Situ.Y=0 then { revient de la ligne zero }
                         begin
                           Em.Saved := FALSE;
                         end;
                     end;
                   Em.Situ.Y := Em.Pro.Par1-64;
                   Em.Situ.X := Em.Pro.Par2-65;
                   { changement de l'att curseur ! }
                   Em.Situ.Att    := AttInit;
                   Em.Situ.Latent := 0;
                   Em.Situ.CurPos := TRUE;  { on connait la pos }
                   CurChange(Em);
                 end; { coord bonne }
               Em.Aig := NUL;
              End;
           end;
       end;
   ESC :
       begin { ESC X [,Y...] }
         case Car of
           #32..#47 : begin { ISO 2022 }
                        Em.Aig := ISO_2022;
                        Em.Pro.Par1 := 0;
                        Em.Pro.Par2 := byte(Car); { on garde cette information }
                      end;
           #53..#55 : Em.Aig := FILT;
           #57..#59 : begin
                        case Car of
                          #57 : Em.Aig := PRO1;
                          #58 : Em.Aig := PRO2;
                          #59 : Em.Aig := PRO3;
                        end;
                        Em.Pro.Par3 := 0;
                      end;
           #64..#71 : { trait }
                    Em.Situ.Att := (Em.Situ.Att and not(Trait))
                                   or CoulPC[byte(Car)-64];
           #72,#73 : { stabilite }
                   Em.Situ.Att := (Em.Situ.Att and not(Stable))
                                  or ((byte(Car)-72) shl 9);
           #76..#79 : { taille }
                    if Em.Situ.Att and Graphik=0 then { texte }
                      Em.Situ.Att := (Em.Situ.Att and not(Taille))
                                     or ((byte(Car)-76) shl 6);
           #80..#87 : { couleur fond }
                      { on indic fond latent }
                    begin
                      Em.Situ.Att := (Em.Situ.Att and not(Fond))
                                     or (CoulPC[byte(Car)-80] shl 3);
                      Em.Situ.Latent := Em.Situ.Latent or Fond;
                    end;
           #88 : { masquage }
                 { on indique masquage latent }
                begin
                  Em.Situ.Att := (Em.Situ.Att and not(Mask)) or Mask;
                  Em.Situ.Latent := Em.Situ.Latent or Mask;
                end;
           #89,#90 : { soulignement/disjointement }
                 begin
                   Em.Situ.Att := (Em.Situ.Att and not(Souligne)) or
                                  ((byte(Car)-89) shl 10);
                   Em.Situ.Latent := Em.Situ.Latent or Souligne;
                 end;
           #91 : begin { CSI }
                   Em.Aig      := ISO_6429; { echappement ISO 6429 }
                   Em.Pro.Par1 := 0;
                   Em.Pro.Par2 := 0;
                   Em.Pro.Par3 := 0;
                 end;
           #92,#93 : { inverse }
                   if Em.Situ.Att and Graphik=0 then { texte }
                     Em.Situ.Att := ((Em.Situ.Att and not(Inverse))
                                    or ((byte(Car)-92) shl 8));
           #95 : { demasquage }
                 { on indique demasquage latent }
                 begin
                   Em.Situ.Att := Em.Situ.Att and not(Mask);
                   Em.Situ.Latent := Em.Situ.Latent or Mask;
                 end;
           #$61 : Emul := EMOP_POSCUR;  { demande position curseur }
         end;
         if Em.Aig=ESC then Em.Aig:=NUL;
         { on en deduit qu'il n'y a pas de suite a donner }
       end;
   FILT : Em.Aig:=NUL; { filtrage }
   PRO1 : begin { PRO1 }
          { l'analyse porte sur Data }
            Case Car Of
             #$68 : Begin { connexion }
                     ResetMinitel(Em);
                     Emul := EMOP_CONNEXION;
                    End;
             #$69 : Begin { deconnexion }
                     ResetMinitel(Em);
                     Emul := EMOP_DECONNEXION;
                    End;
             #$6C : Begin { retournement minitel et emet a 1200 }
                     ResetMinitel(Em);
                     Emul := EMOP_RET;
                    End;
             #$6D : Begin { idem... inverse }
                     ResetMinitel(Em);
                     Emul := EMOP_RETINV;
                    End;
             #$6F : Emul := EMOP_OPPO;
             #$78 : Emul := EMOP_IDEN1;
             #$79 : Emul := EMOP_IDEN2;
             #$7A : Emul := EMOP_ENQRAM1;
             #$7B : Emul := EMOP_ENQROM;   { identification minitel }
             #$7F : Begin { Reset minitel }
                     ResetMinitel(Em);
                     Emul := EMOP_RESET;
                    End;
            End;

            Em.Aig := NUL;
          end;
   PRO2 : begin { PRO2 }
          if Em.Pro.Par3=0 then
            begin
              Em.Pro.Par1 := byte(Car);
              Em.Pro.Par3 := 1;
            end
          else
            begin
              Em.Pro.Par2 := byte(Car);
              { l'analyse porte sur Param1 et Param2 }
              case Em.Pro.Par1 of
                $31 : If Em.Pro.Par2=$7D Then Emul := EMOP_TELEINFO;
                $32 : Case Em.Pro.Par2 Of
                       $7D : Emul := EMOP_MIXTE1;
                       $7E : Emul := EMOP_MIXTE2;
                      End;
                $66 : { TRANSPARENCE }
                      if (Em.Pro.Par2>0) and Em.Cfg.EmulTrans then
                        Em.Aig := TRANS_MODE;
                $69 : Begin { START }
                       Case Em.Pro.Par2 Of
                        $43 : Begin
                               Em.Stat.MdPage := FALSE; { mode rouleau }
                               Emul := EMOP_ROLLON;
                              End;
                        $44 : Emul := EMOP_PROCON;  { Procedure corre Err ON }
                        $45 : Emul := EMOP_MINUSCULE; { passage en minuscule }
                       End;
                      End;
                $6A : Begin { STOP }
                       Case Em.Pro.Par2 Of
                        $43 : Begin
                               Em.Stat.MdPage := TRUE; { mode page }
                               Emul := EMOP_ROLLOFF;
                              End;
                        $44 : Emul := EMOP_PROCOFF; { Fin corre Err }
                        $45 : Emul := EMOP_MAJUSCULE;  { passage en majuscule }
                       End;
                      End;
                $6F : If Em.Pro.Par2=$31 Then Emul := EMOP_OPPORE;
                $73 : Begin  { le status de fonctionnement est }
                             { dans Par2 }
                       Em.Stat.MdPage := Em.Pro.Par2 And 2=0;
                       Emul := EMOP_REPSTATUSFONC;
                      End;
              end;
              { ... }
              if Em.Aig=PRO2 then Em.Aig := NUL;
            end;
        end;
   PRO3 : begin { PRO3 }
          if Em.Pro.Par3>=2 then
            begin
              Em.Pro.Par3 := byte(Car);
              { l'analyse porte sur Par1, Par2 et Par3       }
              { simul gere un ecran minitel virtuel, et donc }
              { ne gere pas le minitel entier!!              }
              { ... }
              Em.Aig := NUL;
            end
          else
            begin
              if Em.Pro.Par3=0 then Em.Pro.Par1 := byte(Car) { si=0 }
              else Em.Pro.Par2 := byte(Car); { si=1 }
              inc(Em.Pro.Par3); { attend le parametre suivant }
            end;
        end;
   ISO_6429 :
        begin { ISO 6429 :  CSI, 3/x...3/z, 2/x...2/z,(64..127) }
          { Par1 = Parametre en cours d'estimation (0..64) }
          { Par2 = Numero Parametre (0..LongPCode-1) }
          { Par3 = etat analyse 0:ok  Autres:2/x...2/z detecte! }
          { Attention : 2/x...2/z annule l'action }
          if (Car>=#48) and (Car<=#57) then
            begin
              if Em.Pro.Par1<25 then Em.Pro.Par1 := Em.Pro.Par1*10+(byte(Car)-48)
              else Em.Pro.Par1 := 250; { signifie le debordement }
            end
          else
            if Car=#59 then
              begin { separateur de parametre detecte }
                Em.Pro.PCode[Em.Pro.Par2] := Em.Pro.Par1;
                Em.Pro.Par1 := 0;
                if Em.Pro.Par2<LongPCode-1 then inc(Em.Pro.Par2);
              end
            else
              if Car>=#64 then
                begin { le top-final de la commande est arrive }
                  if Em.Pro.Par3=0 then { pas de syntaxe error }
                    begin
                      if Em.Situ.Y>0 then
                        begin { on est pas en ligne 0 ! }
                          Em.Pro.PCode[Em.Pro.Par2] := Em.Pro.Par1;
     { PCode[0..Param2] = parametres }
     { il y a obligatoirement 1 parametre ('0' minimun) }
   case Car of
     'A'..'D' : begin { curseur move }
                  Em.Pro.Par1 := Em.Pro.PCode[0];
                    { le 1' param compte }
                  if Em.Pro.Par1=0 then Em.Pro.Par1 := 1;
                    { Param1 = nbre de decalage curseur }
                  case Car of
                    'A' : if Em.Pro.Par1>=Em.Situ.Y then Em.Situ.Y := 1
                          else dec(Em.Situ.Y,Em.Pro.Par1);
                    'B' : begin
                            inc(Em.Situ.Y,Em.Pro.Par1);
                            if Em.Situ.Y>=ScreenHigh then
                              Em.Situ.Y := ScreenHigh-1;
                          end;
                    'C' : begin
                            inc(Em.Situ.X,Em.Pro.Par1);
                            if Em.Situ.X>=ScreenLong then
                              Em.Situ.X := ScreenLong-1;
                          end;
                    'D' : if Em.Pro.Par1>=Em.Situ.X then Em.Situ.X := 0
                          else dec(Em.Situ.X,Em.Pro.Par1);
                  end;
                  CurChange(Em);
                end;

      'J' : begin { effacement niveau ecran }
              Em.Pro.Par1 := Em.Pro.PCode[0]; { le 1' param interesse }
              { Param1 = type d'effacement }
              case Em.Pro.Par1 of
                0 : begin { clr du cur inclus a la fin de l'ecran }
                      ClrLigneMntl(Em,Em.Situ.Y,Em.Situ.X,ScreenLong-1);
                      ClrScrMntl(Em,Em.Situ.Y+1,ScreenHigh-1);
                    end;
                1 : begin { clr du debut ecran a cur inclus }
                      ClrScrMntl(Em,1,Em.Situ.Y-1);
                      ClrLigneMntl(Em,Em.Situ.Y,0,Em.Situ.X);
                    end;
                2 : begin { clr de tout l'ecran , cur inchange }
                      ClrScrMntl(Em,1,ScreenHigh-1);
                    end;
              end;
            end;

      'K' : begin { effacement au niveau d'une ligne }
              Em.Pro.Par1 := Em.Pro.PCode[0]; { le premier code compte }
              { Param1 = type de clr }
              case Em.Pro.Par1 of
                0 : begin { clr du cur inclus a la fin de ligne }
                      ClrLigneMntl(Em,Em.Situ.Y,Em.Situ.X,ScreenLong-1);
                    end;
                1 : begin { clr du debut de ligne au cur inclus }
                      ClrLigneMntl(Em,Em.Situ.Y,0,Em.Situ.X);
                    end;
                2 : begin { clr la ligne , cur inchange }
                      ClrLigneMntl(Em,Em.Situ.Y,0,ScreenLong-1);
                    end;
              end;
            end;

      'L' : begin { Insertion de lignes }
              Em.Pro.Par1 := Em.Pro.PCode[0];
              { Param1 = nbre d'insertion }
              if Em.Pro.Par1 = 0 then Em.Pro.Par1 := 1;
              while Em.Pro.Par1>0 do
                begin
                  ScrDown(Em,Em.Situ.Y);
                  dec(Em.Pro.Par1);
                end;
              Em.Situ.X := 0;
              CurChange(Em);
            end;

      'M' : begin { suppression de lignes }
              Em.Pro.Par1 := Em.Pro.PCode[0];
              { Param1 = nbre de suppression }
              if Em.Pro.Par1=0 then Em.Pro.Par1 := 1;
              while Em.Pro.Par1>0 do
                begin
                  ScrUp(Em,Em.Situ.Y);
                  dec(Em.Pro.Par1);
                end;
              Em.Situ.X := 0;
              CurChange(Em);
            end;

      'P' : begin { suppression de caracteres a partir du cur inclus }
              Em.Pro.Par1 := Em.Pro.PCode[0];
              { Param1 = nbre de suppression }
              if Em.Pro.Par1=0 then Em.Pro.Par1 := 1;
              SupMntl(Em,Em.Situ.X,Em.Situ.Y,Em.Pro.Par1);
            end;

      '@' : if Em.Cfg.RTIC then
              begin { insertion de caracteres a partir du cur inclus }
                Em.Pro.Par1 := Em.Pro.PCode[0];
                { Param1 = nbre d'insertion }
                if Em.Pro.Par1=0 then Em.Pro.Par1 := 1;
                InsMntl(Em,Em.Situ.X,Em.Situ.Y,Em.Pro.Par1);
              end;

      'h' : if Em.Pro.PCode[0]=4 then Em.Stat.MdIns := TRUE;
              { passage en mode insertion }
      'l' : if Em.Pro.PCode[0]=4 then Em.Stat.MdIns := FALSE;
              { passage en mode non-insertion }

      'H' : begin { adressage direct curseur }

              if Em.Pro.Par2=0 then Em.Pro.PCode[1] := 0;
                { second parametre non-initialise }
              Em.Situ.X := byte(Em.Pro.PCode[1]);
              Em.Situ.Y := byte(Em.Pro.PCode[0]);
              if Em.Situ.X>0 then Dec(Em.Situ.X);
              if Em.Situ.Y>1 then Dec(Em.Situ.Y)
              else Em.Situ.Y := 1;
              if Em.Situ.X>ScreenLong-1 then Em.Situ.X := ScreenLong-1;
              if Em.Situ.Y>ScreenHigh-1 then Em.Situ.Y := ScreenHigh-1;
              { on ne met pas Em.Rdy.CurPos=TRUE, car Attr non defini }
              CurChange(Em);
            end;
   end; { case of }
                     Em.Aig := NUL;
                        end { Y>1 }
                      else Em.Aig := NUL; { ignore CSI en ligne 0 }
                    end
                else Em.Pro.Par3 := 1; { indic analyse anormale }
            end; { top of param }
        end; { ISO 6429 }
   ISO_2022 : begin { ISO 2022 : 27, 2/x...2/z, (48..127) }
          { param2 = premier code apres 27 (ESC) }
          Em.Pro.PCode[Em.Pro.Par1] := byte(Car);
          if (Car>=#48) or (Em.Pro.Par1>=LongPCode-1) then
            begin { le top-final est la }
             Case Em.Pro.Par2 Of
              $23 : If (Em.Pro.Par1=1) And (Em.Pro.PCode[0]=$20) Then
                     Case Car Of
                      #$58 : Begin
                              Em.Stat.MskEcran := TRUE;
                              Emul := EMOP_MSKON;
                             End;
                      #$5F : Begin
                              Em.Stat.MskEcran := FALSE;
                              Emul := EMOP_MSKOFF;
                             End;
                     End;
              $28 : If Em.DRCS<>NIL Then
                     Case Em.Pro.Par1 Of
                      0 : If Car=#$40 Then  { Mode G0 }
                           Begin
                            Em.Situ.G0 := TRUE;
                           End;
                      1 : If (Em.Pro.PCode[0]=$20) And (Car=#$42) Then
                           Begin  { Mode G'0 }
                            Em.Situ.G0 := FALSE;
                           End;
                     End;
              $29 : If Em.DRCS<>NIL Then
                     Case Em.Pro.Par1 Of
                      0 : If Car=#$63 Then  { Mode G1 }
                           Begin
                            Em.Situ.G1 := TRUE;
                           End;
                      1 : If (Em.Pro.PCode[0]=$20) And (Car=#$43) Then
                           Begin  { Mode G'1 }
                            Em.Situ.G1 := FALSE;
                           End;
                     End;
             End;

             Em.Aig := NUL;
            end
          else inc(Em.Pro.Par1);
        end; { ISO 2022 }
   TRANS_MODE : begin { emule le mode transparence... }
          dec(Em.Pro.Par2);
          if Em.Pro.Par2=0 then Em.Aig := NUL;
        end;

   JEU_DRCS :   { Definir le n du jeu tlchargeable }
                { on attend: 20 20 JEU 49 }
    If Em.DRCS=NIL Then Em.Aig := NUL
    Else
     Begin
      If Em.Pro.Par3<=1 Then
       Begin
        If Car=' ' Then Inc(Em.Pro.Par3)
        Else Em.Aig := NUL;   { Fin ... }
       End
      Else
       If Em.Pro.Par3=2 Then
        Begin
         If (Car='B') or (Car='C') Then
          Begin
           Em.Pro.Par1 := Byte(Car);  { memorise type de jeu }
           Inc(Em.Pro.Par3);
          End
         Else Em.Aig := NUL;   { Fin ... }
        End
       Else
        Begin
         If Car=#$49 Then
          Begin  { Fin Normal }
           Em.Stat.nG := Em.Pro.Par1-Byte('B');  { nouveau G0 ou G1 }
          End;
         Em.Aig := NUL;
        End;
     End;

   TELE_DRCS :  { telechargement du jeu ... }
                { mais tant que Par1=0, on ignore }
                { Par2=car a telecharger lorsque Par1<>0 }
                { si Car=US alors on sort du tl... }
                { Par3=noctets tlchargs }
    If Em.DRCS=NIL Then Em.Aig := NUL
    Else
     Begin
      If (Em.Pro.Par1<>0) And ((Car>=#$40) And (Car<=#$7F)) Then
       Begin  { Ok ... }
        AccuTele(Em,Car);
       End
      Else
       If Car=#$1F Then
        Begin  { Fin de tl... }
         CompleteTele(Em); { Complete avec des 0 }

         Em.Aig := NUL;
         Emul := EmulCtrl(Em,Car);
        End
       Else
        If Car=#$30 Then
         Begin { Car suivant }
          If Em.Pro.Par1<>0 Then
           Begin { Complete avec des 0 }
            CompleteTele(Em);
           End;

          If Em.Pro.Par2<$7F Then
           Begin
            Em.Pro.Par1 := 1;
            Em.Pro.Par3 := 0;
           End
          Else Em.Pro.Par1 := 0;  { plus prt }
         End;
     End;

 end; { case of }
end;


end.
